from .core import BaseHMF
from .utils import stride_util
from .utils import border_idx_util
from .utils import save_obj
from .utils import load_obj
from .parallel import WriterProcessManager

import numpy as np
import pandas as pd
import os
import shutil
import psutil
from multiprocessing import sharedctypes


GROUPBY_ENCODER = "__specialHMF__groupByNumericEncoder"
MEMMAP_MAP_FILENAME = "__specialHMF__memmapMap"
NUM_FILE_COPY = 3


def open_file(root_path, mode='w+', verbose=False):
    """
    Available modes: 

        w: write only. Truncate if exists.
        w+: read and write. Truncate if exists.
        r: read only.
        r+: read and write. Do not truncate if exists.
    """

    if(mode=='w+'):

        # check if path exists
        # check if it is conforming
        # 
        if(os.path.exists(root_path)):

            shutil.rmtree(root_path)
        
        os.mkdir(root_path)

        from_existing = False
        memmap_map = None

    elif(mode=='r+'):

        # if(os.path.exists(root_path)):

        is_hmf_directory_flag, memmap_map = is_hmf_directory(root_path)

            # if(is_hmf_directory_flag):

            #     from_existing = True

            # else:

            #     from_existing = False

            #     # warn that it is not
            #     # memmap not there
            #     # try to recover...

            #     pass

    hmf = HMF(root_path, memmap_map, verbose)

    return hmf

def is_hmf_directory(root_path):
    """Assuming root_path exists:
    1. check memmap_map exists
    2. check files are all present 

    Needs much improvements...
    """

    # file_list = os.listdir(root_path)

    # if MEMMAP_MAP_FILENAME not in file_list:
    #     print('memmap_map not present')
    #     return(False, None)

    if not fail_safe_check_obj(root_path, MEMMAP_MAP_FILENAME):

        print('memmap_map not present')
        return(False, None)



    memmap_map = fail_safe_load_obj(os.path.join(root_path, MEMMAP_MAP_FILENAME))

    # array_file_list = get_all_array_dirpaths(memmap_map)

    # if(not set(array_file_list) < set(file_list)):

    #     return(False, None)

    return(True, memmap_map)

def get_all_array_dirpaths(m):
    
    visited_dirpath = []
    array_dirpaths = []
    
    _depth_first_search(m, 'HMF_rootNodeKey', visited_dirpath, array_dirpaths)
    
    return array_dirpaths

def _depth_first_search(m, k, visited_dirpaths, array_dirpaths):
    """
    DFS(G, k):
        mark k as visited
        for node l pointed to by node k:
            if node l is not visited:
                DFS(G, l)
                
    * we are marking visits by dirpath not by node name
    * we need a separate list recording array dirpaths
    * when we reach array node, we must stop the search
      because that node does not have [ nodes ] key.
      (or could make an empty nodes dict...)
    
    Parameters
    ----------
    m : dict
        where we can query by k (the node name level map)
        also, the map needs to have been positioned by level
        
    """
    
    if(k == 'HMF_rootNodeKey'):
        cur_dirpath = m['dirpath']
        visited_dirpaths.append(cur_dirpath)
        node_pos = m['nodes']
    
    else:
        cur_dirpath = m[k]['dirpath']
        visited_dirpaths.append(cur_dirpath)
        if(m[k]['node_type']=='array'):
            array_dirpaths.append(cur_dirpath)
            return
        
        node_pos = m[k]['nodes']
    
    for l, v in node_pos.items():
        
        cur_dirpath = node_pos[l]['dirpath']
        
        if(not cur_dirpath in visited_dirpaths):
            _depth_first_search(node_pos, l, visited_dirpaths, array_dirpaths)
    

class HMF(BaseHMF):
    
    def __init__(self, root_dirpath, memmap_map, verbose=True, show_progress=True):
        super(HMF, self).__init__(root_dirpath, memmap_map, verbose)
        
        self.root_dirpath = root_dirpath
        self.arrays = list()
        self.str_arrays = list()
        self.node_attrs = list()

        self.show_progress = show_progress
    
    def from_pandas(self, pdf, groupby=None, orderby=None):
        """
        need to numerify groupby col!"""
        
        self.pdf = pdf
        
        if groupby and orderby:
            self.pdf[GROUPBY_ENCODER] = self.pdf[groupby].astype('category')
            self.pdf[GROUPBY_ENCODER] = self.pdf[GROUPBY_ENCODER].cat.codes

            self.pdf = self.pdf.sort_values(by=[groupby, orderby]).reset_index(drop=True)
            group_array = self.pdf[GROUPBY_ENCODER].values

            tmp = pd.DataFrame(self.pdf[groupby].unique(), columns=[groupby])
            tmp = tmp.sort_values(by=groupby).reset_index(drop=True)
            group_names = tmp[groupby].tolist()
            
        elif orderby:
            self.pdf = self.pdf.sort_values(by=[orderby]).reset_index(drop=True)

            group_array = np.zeros(len(pdf))
            group_names = ['0']
            
        elif groupby:
            self.pdf[GROUPBY_ENCODER] = self.pdf[groupby].astype('category')
            self.pdf[GROUPBY_ENCODER] = self.pdf[GROUPBY_ENCODER].cat.codes

            self.pdf = self.pdf.sort_values(by=[groupby]).reset_index(drop=True)
            group_array = self.pdf[GROUPBY_ENCODER].values

            tmp = pd.DataFrame(pdf[groupby].unique(), columns=[groupby])
            tmp = tmp.sort_values(by=groupby).reset_index(drop=True)
            group_names = tmp[groupby].tolist()
            
        else:
            group_array = np.zeros(len(pdf))
            group_names = ['0']
            
        border_idx = border_idx_util(group_array)
        group_idx = stride_util(border_idx, 2, 1, np.int32)

        self.group_sizes = np.diff(border_idx)
        self.group_names = group_names
        self.groups = list(zip(group_names, group_idx))

    def register_array(self, array_filename, col_names, encoder=None, decoder=None):
        """Update memmap_map dictionary - which assumes all saves will be successful.
        We need to validity check on arrays
        Also put arrays into sharedctypes
        """

        if(encoder):
            data_array = encoder(self.pdf[col_names])
        else:
            data_array = self.pdf[col_names].values
            
        
        self.arrays.append((array_filename, data_array))

    def register_node_attr(self, attr_dirpath, key, value):

        self.node_attrs.append((attr_dirpath, key, value))




    def _write_registered_node_attrs(self):
        """
        The logic used in this method largely mirrors those found in parallel.py.

        Main difference: 
            1. no need to parallelize this
            2. we can rely on baseHMF for this since we have access to the HMF object
        """

        tasks = list(itertools.product(
            range(len(self.node_attrs)), 
            range(len(self.groups))))

        for task in tasks:

            attr_dirpath_standalone = self.node_attrs[task[0]][0]
            key_standalone = self.node_attrs[task[0]][1]
            value_standalone = self.node_attrs[task[0]][2]

            group_name = self.groups[task[1]][0]

            print(group_name, attr_dirpath_standalone, key_standalone, value_standalone)



        # array_filename = self.hmf_obj.arrays[task[0]][0]
        # shared_array = self.hmf_obj.arrays[task[0]][1]

        # group_name = self.hmf_obj.groups[task[1]][0]
        # start_idx, end_idx = self.hmf_obj.groups[task[1]][1]

        
        # if(len(self.hmf_obj.groups)==1):
        #     array_filepath = '/'.join((self.hmf_obj.root_dirpath, array_filename))
        # else:
        #     array_filename = self.hmf_obj._assemble_dirpath(group_name, array_filename)
        #     array_filepath = '/'.join((self.hmf_obj.root_dirpath, array_filename))

        
    def close(self, zip_file=False, num_subprocs='auto'):
        """
        How we process the str_arrays should depend on how many arrays we have VS how many
        subprocs we can open
        """



        if(len(self.arrays) > 0):

            if(num_subprocs=='auto'):
                num_subprocs = psutil.cpu_count(logical=False) - 1

            if(self.verbose):
                print('Saving registered arrays using multiprocessing [ {} ] subprocs\n'.format(num_subprocs))

            WPM = WriterProcessManager(self, num_subprocs=num_subprocs, verbose=self.verbose)
            WPM.start()


            self.failed_tasks = WPM.failed_tasks

        memmap_map_dirpath = os.path.join(self.root_dirpath, MEMMAP_MAP_FILENAME)

        fail_safe_save_obj(self.memmap_map, memmap_map_dirpath)


# PR 0.0.b16
def fail_safe_save_obj(obj, dirpath):

    for i in range(NUM_FILE_COPY):

        try:

            copy_dirpath = dirpath + str(i)

            save_obj(obj, copy_dirpath)

        except:

            continue

def fail_safe_load_obj(dirpath):

    for i in range(NUM_FILE_COPY):

        try:

            copy_dirpath = dirpath + str(i)

            return load_obj(copy_dirpath)

        except:

            continue
            
    raise IOError("Damn it, failed to read file again")

def fail_safe_check_obj(root_path, filename):

    file_list = os.listdir(root_path)

    for i in range(NUM_FILE_COPY):

        copy_filename = filename + str(i)

        if copy_filename in file_list:

            return True




            


from .utils import stride_util
from .utils import border_idx_util
from .utils import write_memmap
from .utils import read_memmap
from .utils import printProgressBar

from multiprocessing import Process, Manager, sharedctypes
import itertools
import numpy as np
from collections import defaultdict
import time


MAX_WRITE_ATTEMPT = 3
READ_WAIT_INTERVALS = [0.1, 1.0, 3.0]
MAX_READ_ATTEMPT = len(READ_WAIT_INTERVALS)


class WriterProcess(Process):
    
    def __init__(self, shared_write_result_dict, shared_write_error_dict, task):
        super(WriterProcess, self).__init__()

        self.shared_write_result_dict = shared_write_result_dict
        self.shared_write_error_dict = shared_write_error_dict
        
        self.task_key = task

        array_filename = SHARED_HMF_OBJ.arrays[task[0]][0]
        

        group_name = SHARED_HMF_OBJ.groups[task[1]][0]
        start_idx, end_idx = SHARED_HMF_OBJ.groups[task[1]][1]


        self.array = SHARED_HMF_OBJ.arrays[task[0]][1][start_idx:end_idx]

        
        if(len(SHARED_HMF_OBJ.groups)==1):
            self.array_filepath = '/'.join((SHARED_HMF_OBJ.root_dirpath, array_filename))
        else:
            array_filename = SHARED_HMF_OBJ._assemble_dirpath(group_name, array_filename)
            self.array_filepath = '/'.join((SHARED_HMF_OBJ.root_dirpath, array_filename))
        
    def run(self):
        """
        We need a way to surface the failure reasons.
        In case of actual failed task.
        """
        
        # self.write_array(self.array_filepath, self.array)
        
        try:
            # start_time = time.time()
            self.write_array(self.array_filepath, self.array)

            # print(time.time() - start_time)

            self.shared_write_result_dict[self.task_key] = 'success'
            
        except Exception as e:
            
            self.shared_write_error_dict[self.task_key] = str(e)
            self.shared_write_result_dict[self.task_key] = 'failure'
    
    def write_array(self, array_filepath, array):
        
        dtype = str(array.dtype)
        shape = array.shape
        write_memmap(array_filepath, dtype, shape, array)


class WriterProcessManager():
    
    def __init__(self, hmf_obj, num_subprocs=4, verbose=True, show_progress=True):


        # start_time = time.time()
        global SHARED_HMF_OBJ
        SHARED_HMF_OBJ = hmf_obj

        self.hmf_obj = hmf_obj
        
        self.tasks = list(itertools.product(
            range(len(hmf_obj.arrays)), 
            range(len(hmf_obj.groups))))
        self.pending_tasks = []
        self.failed_tasks = []
        self.successful_tasks = []
        
        self.write_attempt_dict = defaultdict(int)
        self.read_attempt_dict = defaultdict(int)

        self.successful_write_tasks = []
        self.failed_write_tasks = []

        self.successful_read_tasks = []
        self.failed_read_tasks = []
        
        manager = Manager()
        self.shared_write_result_dict = manager.dict()
        self.shared_write_error_dict = manager.dict()
        self.shared_read_error_dict = dict()
        self.subprocs = []
        
        self.num_subprocs = num_subprocs
        self.verbose = verbose

        self.max_len = len(self.tasks)
        self.show_progress = show_progress
        
    def read_task(self, task):
    
        array_filename = self.hmf_obj.arrays[task[0]][0]
        shared_array = self.hmf_obj.arrays[task[0]][1]

        group_name = self.hmf_obj.groups[task[1]][0]
        start_idx, end_idx = self.hmf_obj.groups[task[1]][1]

        # array_filename = self.hmf_obj._assemble_dirpath(group_name, array_filename)
        # array_filepath = '/'.join(('', group_name, array_filename))


        if(len(self.hmf_obj.groups)==1):
            array_filepath = '/'.join(('', array_filename))
        else:
            array_filepath = '/'.join(('', group_name, array_filename))



        try: 
            self.hmf_obj.get_array(array_filepath)
            return('success')

        except Exception as e:
            self.shared_read_error_dict[task] = str(e)
            return('failure')


    def write_task(self, task):
        """Update memmap_map here.

        This logic only supports single level group...
        """

        subproc = WriterProcess(self.shared_write_result_dict,self.shared_write_error_dict,
            task)



        subproc.daemon = True
        subproc.start()
        return subproc

    def update_memmap_map(self, task):

        array_filename = self.hmf_obj.arrays[task[0]][0]
        shared_array = self.hmf_obj.arrays[task[0]][1]

        group_name = self.hmf_obj.groups[task[1]][0]
        start_idx, end_idx = self.hmf_obj.groups[task[1]][1]


        # array_filename = self.hmf_obj._assemble_dirpath(group_name, array_filename)
        array = np.ctypeslib.as_array(shared_array)[start_idx:end_idx]

        if(len(self.hmf_obj.groups)==1):
            array_filepath = '/'.join(('', array_filename))
        else:
            array_filepath = '/'.join(('', group_name, array_filename))

        

        self.hmf_obj.update_memmap_map_array(array_filepath, array)


    
    def start(self):

        if(self.show_progress):
            printProgressBar(0, self.max_len, prefix = 'Progress:', suffix = '', length = 50)
        
        # pending_tasks: written and either not read or failed to read < MAX_READ_ATTEMPT times
        # tasks: not yet written or failed to read MAX_READ_ATTEMPT times
        while(len(self.pending_tasks) > 0 or len(self.tasks) > 0):

            # if we can run more subprocs and if there are tasks remaining
            # run a task in subproc
            if(len(self.subprocs) < self.num_subprocs and len(self.tasks) > 0):

                # this task will not be put back for writing again 
                # until read is failed MAX_READ_ATTEMPT times
                task = self.tasks.pop()

                if self.verbose:
                    print('#######################################\n')
                    print('Working on {}'.format(task))

                # it is pending until read is successful or failed MAX_READ_ATTEMPT times
                # if read is successful, move this task to successful_tasks
                # if read is failed MAX_READ_ATTEMPT times, move it back to tasks
                # --> will count up the write attempt by 1
                # if write count reaches MAX_WRITE_ATTEMPT, move this task to failed_tasks
                # so we can have failures without crashing the program
                self.pending_tasks.append(task)

                subproc = self.write_task(task)
                self.write_attempt_dict[task] += 1

                self.subprocs.append(subproc)

            if self.verbose:
                print('\nCHECKING WRITE STATUS')
                print('from shared_write_result_dict: {}\n'.format(self.shared_write_result_dict))

            if(len(self.shared_write_result_dict.keys()) > 0):

                for tried_write_task in self.shared_write_result_dict.keys():

                    if self.shared_write_result_dict[tried_write_task] == 'success':

                        if self.verbose:
                            print('Write successful: {}'.format(tried_write_task))
                        self.successful_write_tasks.append(tried_write_task)

                        if self.verbose:
                            print('Dropping {} from shared dict for success'.format(tried_write_task))
                        self.shared_write_result_dict.pop(tried_write_task, None)

                        if self.verbose:
                            print('Updating memmap map')
                        self.update_memmap_map(tried_write_task)

                    elif self.shared_write_result_dict[tried_write_task] == 'failure':

                        if self.verbose:
                            print('Write failed: {}'.format(tried_write_task))

                        # we need to push these back to tasks
                        # since we need to start the process again, 
                        # no need for any wait time 
                        # (there is nothing to wait for! it failed!)
                        self.failed_write_tasks.append(tried_write_task)

                        if self.verbose:
                            print('Dropping {} from shared dict for failure'.format(tried_write_task))
                        self.shared_write_result_dict.pop(tried_write_task, None)


            if self.verbose:
                print('\nCHECKING READ STATUS')
                print('from successful_write_tasks: {}\n'.format(self.successful_write_tasks))
            # we record the failed read tasks here
            # the purpose is to temporarily hold these 
            # and escape while loop over successful write tasks
            # add them back in later

            # these are tasks with successful writes
            # but fails at the read
            # what we want for these:
            # need to try reading some more
            self.failed_read_tasks = []

            while(len(self.successful_write_tasks) > 0):

                successful_write_task = self.successful_write_tasks.pop()

                # read_result = self.read_task(successful_write_task)
                read_result = 'success'
                self.read_attempt_dict[successful_write_task] += 1

                if read_result == 'success':

                    if self.verbose:
                        print('Read successful: {}'.format(successful_write_task))

                    self.successful_read_tasks.append(successful_write_task)

                elif read_result == 'failure':

                    if self.verbose:
                        print('Read failed: {}'.format(successful_write_task))

                    # since the read was not successful for this written task, 
                    # put the written task back into the tasks list:
                    self.failed_read_tasks.append(successful_write_task)

                    read_attempt_cnt = self.read_attempt_dict[successful_write_task] - 1
                    time.sleep(READ_WAIT_INTERVALS[read_attempt_cnt])

            # throw the failed read tasks back in 
            # for these, we don't want to try to write again
            # we want to read again
            self.successful_write_tasks += self.failed_read_tasks
            self.successful_write_tasks = list(set(self.successful_write_tasks))

            if self.verbose:
                print('\n-------')
                print('successful_write_tasks: ', self.successful_write_tasks)
                print('successful_read_tasks: ', self.successful_read_tasks)

                print('write_attempt_dict: ', self.write_attempt_dict)
                print('read_attempt_dict: ', self.read_attempt_dict)

                # print('shared_write_result_dict', )

            # update pending, successful, failed tasks
            if self.verbose:
                print('\nUPDATE PENDINGS')
                print('pending tasks: {}\n'.format(self.pending_tasks))

            if(len(self.pending_tasks) > 0):

                for _ in range(len(self.pending_tasks)):

                    pending_task = self.pending_tasks.pop()

                    if self.verbose:
                        print('Looking at pending task {}'.format(pending_task))

                    if pending_task in self.failed_write_tasks:

                        if self.verbose:
                            print('    pending --> tasks: {}'.format(pending_task))
                            print('    * write failed\n')

                        # don't append since that will try to do this again right away
                        self.tasks.insert(0, pending_task)
                        self.failed_write_tasks.remove(pending_task)

                    elif pending_task in self.successful_read_tasks:

                        if self.verbose:
                            print('    pending --> successful tasks: {}\n'.format(pending_task))

                        self.successful_tasks.append(pending_task)

                    elif(self.write_attempt_dict[pending_task] == MAX_WRITE_ATTEMPT 
                        and self.read_attempt_dict[pending_task] == MAX_READ_ATTEMPT):

                        if self.verbose:
                            print('    pending --> failed tasks: {}'.format(pending_task))
                            print('    * reached MAX_WRITE_ATTEMPT\n')

                        self.failed_tasks.append(pending_task)
                        self.successful_write_tasks.remove(pending_task)

                    elif self.read_attempt_dict[pending_task] == MAX_READ_ATTEMPT:

                        if self.verbose:
                            print('    pending --> tasks: {}'.format(pending_task))
                            print('    * reached MAX_READ_ATTEMPT\n')

                        self.tasks.insert(0, pending_task)
                        self.read_attempt_dict[pending_task] = 0
                        self.successful_write_tasks.remove(pending_task)

                    else:

                        if self.verbose:
                            print('    nothing to update\n')

                        self.pending_tasks.insert(0, pending_task)

                    if(self.show_progress):



                        # cur_len = int((len(self.successful_tasks) + len(self.failed_tasks))/self.max_len * 100) 
                        cur_len = int(len(self.successful_tasks) + len(self.failed_tasks)) 

                        # print(cur_len/self.max_len, cur_len, self.max_len)


                        printProgressBar(cur_len, self.max_len, prefix = 'Progress:', suffix = '', length = 50)
                    

            # check status of subprocs after all the above is finished
            # to give subprocs more time to write
            self.subprocs = [elem for elem in self.subprocs if elem.is_alive()]

            if self.verbose:
                print('SUBPROC STATUS')
                print('number of outstanding tasks: {} ( {} )'.format(len(self.tasks), self.tasks))
                print('number of running procs: {}'.format(len(self.subprocs))); print()

            # if all procs are busy, there is no gain from trying to 
            # assign work to them at the moment
            if(len(self.subprocs) == self.num_subprocs):
                time.sleep(0.01)

            # if number of tasks is 0, there is no gain from trying to
            # assign work to them at the moment
            # also, if subproc is still running, sleep to give it time to finish
            if(len(self.subprocs) > 0 and len(self.tasks) == 0):
                time.sleep(0.01)

        self.failure_reasons = dict(self.shared_write_error_dict)

        if(self.show_progress):
            printProgressBar(self.max_len, self.max_len, prefix = 'Progress:', suffix = 'Completed!', length = 50)
            print()






__version__ = "0.0.b20"

from .core import BaseHMF
from .parallel import WriterProcessManager
from .hmf import HMF
from .hmf import open_file
from .hmf import is_hmf_directory
from .utils import write_memmap
from .utils import read_memmap
	
__all__ = [
	"BaseHMF",
	"WriterProcessManager",
	"HMF",
	"open_file",
	"is_hmf_directory"
	]
	


from .utils import write_memmap
from .utils import read_memmap
from .utils import load_obj

import os


class BaseHMF():
    
    def __init__(self, root_dirpath, memmap_map, verbose):
        """If root_dirpath exists, load the existing memmap so we can get arrays.
        
        (root)

        attributes
        dirpath
        node_type = root
        nodes
        
            group_name1

                attributes
                dirpath
                node_type = group
                nodes

                    array_name1

                        attributes
                        dirpath
                        node_type = array
                        dtype
                        shape

            group_name2

            group_name3 ...

        Question: why not combine the node name and the dict following it into a single dict?
        - this will in fact make searching thru nodes at nodes key much harder since we can't 
          query it using key name, but entire dictionaries will come up --> forcing us to 
          query using a dictionary.
        
        """
        
        if(memmap_map):

            if(verbose):

                print('Loading from existing HMF')

            self.memmap_map = memmap_map

        else:

            if(verbose):

                print('Creating new HMF')

            self.memmap_map = dict()
            self.memmap_map['attributes'] = dict()
            self.memmap_map['nodes'] = dict()
            self.memmap_map['dirpath'] = root_dirpath
            self.memmap_map['node_type'] = 'root'

        self.verbose = verbose
    
    def set_group(self, group_dirpath):
        """
        Todo: validate group_dirpath
        """
        
        self.update_memmap_map_group(group_dirpath)

        
    def update_memmap_map_group(self, group_dirpath):
        
        memmap_map_pos = self.memmap_map['nodes']
        # dict with node_names of group names / array names

        if(group_dirpath=='/'):
            return 

        # iterate through each level of group (horizontal traversal)
        for idx, node_name in enumerate(group_dirpath.split('/')):

            # skip / because root is not part of dict ( (root) )
            if idx==0:

                if(node_name!=''):
                    raise ValueError('Path much start with the root directory "/"')

                continue    

            # if node_name is one of memmap_map_pos keys
            if(node_name in memmap_map_pos):

                if self.verbose:
                    print('{} found in memmap'.format(node_name))

                if(not self._is_group_node(memmap_map_pos, node_name)):

                    if self.verbose:
                        print('{} is not a group node so skipping'.format(node_name))

                    continue

                # take the current level dirpath
                group_dirpath_pos = memmap_map_pos[node_name]['dirpath']

                # move to the groups dict position of the current node_name
                memmap_map_pos = memmap_map_pos[node_name]['nodes']
                
            else:

                if self.verbose:
                    print('{} not found in memmap'.format(node_name))
                
                # if node_name is not part of the key, create the node_name's map position
                memmap_map_pos[node_name] = dict()
                memmap_map_pos[node_name]['nodes'] = dict()  # create node names dict
                memmap_map_pos[node_name]['attributes'] = dict()
                memmap_map_pos[node_name]['node_type'] = 'group'

                # if it's a node_name right after / root position
                # will always visit - setting the group_dirpath_pos always
                if idx==1:

                    # start the dirpath afresh
                    memmap_map_pos[node_name]['dirpath'] = node_name
                    group_dirpath_pos = memmap_map_pos[node_name]['dirpath']

                else:

                    # if it's beyond the level right after root /
                    # the dirpath is combination of all previous node_name dirpaths
                    memmap_map_pos[node_name]['dirpath'] = self._assemble_dirpath(group_dirpath_pos, node_name)

                # update the map position before moving to the next level of node_name (if there is any)
                memmap_map_pos = memmap_map_pos[node_name]['nodes']
                
    def retrieve_memmap_map_pos_group(self, group_dirpath):
        """
        group_pos: the dictionary position where nodes, dirpath, node_type can be queried
        i.e. right below the node names dict level


        check if it is indeed a group

        """

        if(group_dirpath=='/'):
            return self.memmap_map

        memmap_map_pos = self.memmap_map['nodes']
        
        for idx, node_name in enumerate(group_dirpath.split('/')):
            
            if idx==0:
                continue
            
            if node_name in memmap_map_pos:
            
                memmap_map_group_pos = memmap_map_pos[node_name]
                memmap_map_pos = memmap_map_group_pos['nodes']

            else:
                raise 
                
        return memmap_map_group_pos

    def set_array(self, array_filepath, array):
        
        # validate array
        # validate dirpath
        
        self.update_memmap_map_array(array_filepath, array)
        
        self.write_array(array_filepath, array)
        
    def update_memmap_map_array(self, array_filepath, array):
        
        array_filepaths = array_filepath.split('/')
        
        group_dirpath = '/'.join(array_filepaths[0:-1])
        array_name = array_filepaths[-1]
        
        if len(array_filepaths) > 2:
            self.update_memmap_map_group(group_dirpath)
            
            memmap_map_group_pos = self.retrieve_memmap_map_pos_group(group_dirpath)
        else:
            memmap_map_group_pos = self.memmap_map


        # TODO: check if array_name is already in here
        
        memmap_map_group_pos['nodes'][array_name] = dict()
        memmap_map_group_pos['nodes'][array_name]['node_type'] = 'array'
        memmap_map_group_pos['nodes'][array_name]['attributes'] = dict()
        memmap_map_group_pos['nodes'][array_name]['nodes'] = dict()

        if(memmap_map_group_pos['node_type']!='root'):

            filepath = self._assemble_dirpath(memmap_map_group_pos['dirpath'], array_name)

        else:

            filepath = array_name


        memmap_map_group_pos['nodes'][array_name]['dirpath'] = filepath
        
        dtype = str(array.dtype)
        shape = array.shape
        
        memmap_map_group_pos['nodes'][array_name]['dtype'] = dtype
        memmap_map_group_pos['nodes'][array_name]['shape'] = shape

    def retrieve_memmap_map_pos_array(self, array_filepath):
        
        array_filepaths = array_filepath.split('/')
        
        group_dirpath = '/'.join(array_filepaths[0:-1])
        array_name = array_filepaths[-1]
        
        if len(array_filepaths) > 2:
            self.update_memmap_map_group(group_dirpath)
            
            memmap_map_group_pos = self.retrieve_memmap_map_pos_group(group_dirpath)
        else:
            memmap_map_group_pos = self.memmap_map
            
        return memmap_map_group_pos['nodes'][array_name]

    def write_array(self, array_filepath, array):
        
        memmap_map_array_pos = self.retrieve_memmap_map_pos_array(array_filepath)
        
        dtype = memmap_map_array_pos['dtype']
        shape = memmap_map_array_pos['shape']
        filepath = os.path.join(
            self.memmap_map['dirpath'],
            memmap_map_array_pos['dirpath'])
        
        write_memmap(filepath, dtype, shape, array)

    def get_array(self, array_filepath, idx=None):
        
        memmap_map_array_pos = self.retrieve_memmap_map_pos_array(array_filepath)
        
        dtype = memmap_map_array_pos['dtype']
        shape = memmap_map_array_pos['shape']
        filepath = os.path.join(
            self.memmap_map['dirpath'],
            memmap_map_array_pos['dirpath'])
        
        return read_memmap(filepath, dtype, shape, idx)

    def set_node_attr(self, attr_dirpath, key, value):

        self.update_memmap_map_group(attr_dirpath)
            
        memmap_map_group_pos = self.retrieve_memmap_map_pos_group(attr_dirpath)

        memmap_map_group_pos['attributes'][key] = value

    def get_node_attr(self, attr_dirpath, key):
        
        memmap_map_group_pos = self.retrieve_memmap_map_pos_group(attr_dirpath)

        return memmap_map_group_pos['attributes'][key]

    def _assemble_dirpath(self, source_path, dest_path):

        return '__'.join((source_path, dest_path))

    def _is_group_node(self, memmap_map_pos, node_name):

        return memmap_map_pos[node_name]['node_type'] == 'group'

    def _is_array_node(self, memmap_map_pos, node_name):

        return memmap_map_pos[node_name]['node_type'] == 'array'      
    


import numpy as np
import pickle
import functools
import time
import os

def save_obj(obj, filepath):
    with open(filepath, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        
def load_obj(filepath):
    # with open(filepath, 'rb') as f:
    #     return pickle.load(f)


    f = os.open(filepath, os.O_RDONLY|os.O_NONBLOCK)

    readable = os.fdopen(f, "rb", 0)

    if select.select([readable], [], [], 10)[0][0] == readable:

        return pickle.load(readable)

def stride_util(array, window_size, skip_size, dtype):
    
    return(np.lib.stride_tricks.as_strided(
        array, 
        shape=[np.ceil(len(array)-1).astype(dtype), window_size],
        strides=[array.strides[0]*skip_size, array.strides[0]]))

def border_idx_util(array):
    
    border_idx = np.where(np.diff(array) != 0)[0] + 1
    border_idx = border_idx.astype(np.int)

    border_idx = np.insert(border_idx, 0, 0, axis=0)
    border_idx = np.append(border_idx, [len(array)], axis=0)
    
    return border_idx

def write_memmap(filepath, dtype, shape, array):

    writable_memmap = np.memmap(filepath, dtype=dtype, mode="w+", shape=shape)
    writable_memmap[:] = array[:]
    del writable_memmap

def read_memmap(filepath, dtype, shape, idx=None):

    readonly_memmap = np.memmap(filepath, dtype=dtype, mode="r", shape=shape)

    if idx is None:
        array = readonly_memmap[:]
    else:
        if is_nested_list(idx):
            array = readonly_memmap[tuple(idx)]
        else:
            array = readonly_memmap[idx]

    del readonly_memmap
    return array

def is_nested_list(l):
    
    if not isinstance(l, list):
        return False
    else:
        return any(isinstance(i, list) for i in l)

def printProgressBar (iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)
    # Print New Line on Complete
    if iteration == total: 
        print(f'\r', end = printEnd)

